#!/usr/bin/env python

import sys
sys.path.append('.')
sys.path.append('..')

import os
import argparse

import numpy as np
from extra_data import open_run, by_id, RunDirectory
from pes_to_spec.model import Model, matching_ids

# for helper plots
from pes_to_spec.model import SelectRelevantLowResolution
from pes_to_spec.model import HighResolutionSmoother
from sklearn.decomposition import PCA

from itertools import product

import pandas as pd
from copy import deepcopy
import scipy
from scipy.signal import fftconvolve

from typing import Dict, Optional

from time import time_ns
import pandas as pd

def get_gas(run, tids):
    gas_sources = [
                "SA3_XTD10_PES/DCTRL/V30300S_NITROGEN",
                "SA3_XTD10_PES/DCTRL/V30310S_NEON",
                "SA3_XTD10_PES/DCTRL/V30320S_KRYPTON",
                "SA3_XTD10_PES/DCTRL/V30330S_XENON",
            ]
    gas_active = list()
    for gas in gas_sources:
        # check if this gas source is interlocked
        if gas in run.all_sources and run[gas, "interlock.AActionState.value"].ndarray().sum() == 0:
            # it is not, so this gas was used
            gas_active += [gas.split("/")[-1].split("_")[-1]]
    gas = "_".join(gas_active)
    return gas


def save_result(filename: str,
                spec_pred: Dict[str, np.ndarray],
                spec: np.ndarray,
                spec_smooth: np.ndarray,
                spec_raw_pe: np.ndarray,
                intensity: float,
                pes: Optional[np.ndarray]=None,
                first: Optional[int]=None,
                last: Optional[int]=None,
                ):
    """
    Plot result with uncertainty band.

    Args:
      filename: Output file name.
      spec_pred: Predicted result with uncertainty bands in a dictionary.
      spec: The unsmoothed spectrum with shape (features,).
      spec_smooth: Smoothened expected result with shape (features,).
      spec_raw_pe: x axis with the photon energy in eV.
      pes: PES spectrum for the inset.
      intensity: The XGM intensity in uJ.

    """
    unc_stat = spec_pred["unc"]
    unc_pca = spec_pred["pca"]
    unc = np.sqrt(unc_stat**2 + unc_pca**2)
    df = pd.DataFrame(dict(energy=spec_raw_pe,
                           spec=spec,
                           spec_smooth=spec_smooth,
                           prediction=spec_pred["expected"],
                           unc=unc,
                           unc_pca=unc_pca,
                           unc_stat=unc_stat,
                           beam_intensity=intensity*1e-3*np.ones_like(spec_raw_pe),
                           deconvolved=spec_pred["deconvolved"]
                           ))
    df.to_csv(filename)
    if pes is not None:
        pes_data = deepcopy(pes)
        pes_data['bin'] = np.arange(len(pes['channel_1_D']))
        pes_data['first'] = first*np.ones_like(pes_data['bin'])
        pes_data['last'] = last*np.ones_like(pes_data['bin'])
        df = pd.DataFrame(pes_data)
        df.to_csv(filename.replace('.pdf', '_pes.csv'))

def save_pes_result(filename: str,
                pes: Optional[np.ndarray]=None,
                first: Optional[int]=None,
                last: Optional[int]=None,
                ):
    """
    Plot result with uncertainty band.

    Args:
      filename: Output file name.
      pes: PES spectrum for the inset.
      first: Start of ROI.
      last: End of ROI.

    """
    pes_data = deepcopy(pes)
    pes_data['bin'] = np.arange(len(pes['channel_1_D']))
    pes_data['first'] = first*np.ones_like(pes_data['bin'])
    pes_data['last'] = last*np.ones_like(pes_data['bin'])
    df = pd.DataFrame(pes_data)
    df.to_csv(filename)

def main():
    """
    Main entry point. Reads some data, trains and predicts.
    """
    parser = argparse.ArgumentParser(prog="offline_analysis", description="Test pes2spec doing an offline analysis of the data.")
    parser.add_argument('-p', '--proposal', type=int, metavar='INT', help='Proposal number', default=2828)
    parser.add_argument('-r', '--run', type=str, metavar='INT,INT,...', help='Run numbers, comma-separated.', default=206)
    parser.add_argument('-t', '--test-run', type=int, metavar='INT', help='Run to test', default=None)
    parser.add_argument('-m', '--model', type=str, metavar='FILENAME', default="", help='Model to load. If given, do not train a model and just do inference with this one.')
    parser.add_argument('-d', '--directory', type=str, metavar='DIRECTORY', default=".", help='Where to save the results.')
    parser.add_argument('-S', '--spec', type=str, metavar='NAME', default="SA3_XTD10_SPECT/MDL/SPECTROMETER_SCS_NAVITAR:output", help='SPEC name')
    parser.add_argument('-P', '--pes', type=str, metavar='NAME', default="SA3_XTD10_PES/ADC/1:network", help='PES name')
    parser.add_argument('-X', '--xgm', type=str, metavar='NAME', default="SA3_XTD10_XGM/XGM/DOOCS:output", help='XGM name')
    parser.add_argument('-o', '--offset', type=int, metavar='INT', default=0, help='Train ID offset')
    parser.add_argument('-c', '--xgm_cut', type=float, metavar='INTENSITY', default=0, help='XGM intensity threshold in uJ.')
    parser.add_argument('-T', '--model-type', type=str, metavar='TYPE', default="ard", choices=["bnn", "bnn_rvm", "ridge", "ard"], help='Which model type to use.')
    parser.add_argument('-w', '--weight', action="store_true", default=False, help='Whether to reweight data as a function of the pulse energy to make it invariant to that.')

    args = parser.parse_args()

    print("Opening run ...")
    runs = args.run.split(',')
    runs = [int(r) for r in runs]
    # get run
    run_list = [open_run(proposal=args.proposal, run=r) for r in runs]
    run = run_list[0]
    if len(run_list) > 1:
        run = run.union(*run_list[1:])

    run_test = run
    other_run_test = False
    if "test_run" in args and args.test_run is not None:
        other_run_test = True
        run_test = open_run(proposal=args.proposal, run=args.test_run)

    #run = RunDirectory("/gpfs/exfel/data/scratch/tmichela/data/r0206")

    # ----------------Used in the first tests-------------------------
    # get train IDs and match them, so we are sure to have information from all needed sources
    # in this example, there is an offset of -2 in the SPEC train ID, so correct for it
    spec_offset = -2
    spec_name = 'SA3_XTD10_SPECT/MDL/FEL_BEAM_SPECTROMETER_SQS1:output'
    pes_name = 'SA3_XTD10_PES/ADC/1:network'

    spec_offset = 0
    spec_name = 'SA3_XTD10_SPECT/MDL/SPECTROMETER_SQS_NAVITAR:output'
    pes_name = 'SA3_XTD10_PES/ADC/1:network'
    # -------------------End of test setup ----------------------------

    spec_offset = args.offset
    spec_name = args.spec
    pes_name = args.pes
    xgm_name = args.xgm

    pes_tid = run[pes_name, "digitizers.trainId"].ndarray()
    xgm_tid = run[xgm_name, "data.trainId"].ndarray()

    if len(args.model) == 0:
        spec_tid = spec_offset + run[spec_name, "data.trainId"].ndarray()
        # these are the train ID intersection
        # this could have been done by a select call in the RunDirectory, but it would not correct for the spec_offset
        tids = matching_ids(spec_tid, pes_tid, xgm_tid)

        # read the spec photon energy and intensity
        spec_raw_pe = run[spec_name, "data.photonEnergy"].select_trains(by_id[tids - spec_offset]).ndarray()
        spec_raw_int = run[spec_name, "data.intensityDistribution"].select_trains(by_id[tids - spec_offset]).ndarray()

    else: # when doing inference, no need to load SPEC data
        tids = pes_tid

    # reserve part of it for the test stage
    train_tids = tids[:-10]
    if other_run_test:
        pes_tidt = run_test[pes_name, "digitizers.trainId"].ndarray()
        xgm_tidt = run_test[xgm_name, "data.trainId"].ndarray()
        spec_tidt = run_test[spec_name, "data.trainId"].ndarray()
        test_tids = matching_ids(spec_tidt, pes_tidt, xgm_tidt)
    else:
        test_tids = tids
    print(f"Number of train IDs: {len(train_tids)}")
    print(f"Number of test IDs: {len(test_tids)}")

    # read the PES data for each channel
    #channels = [f"channel_{i}_{l}"
    #            for i, l in product(range(1, 5), ["A", "B", "C", "D"])]
    channels = [f"channel_{i}_{l}"
                for i, l in product([1,3,4], ["A", "B", "C", "D"])]
    pes_raw = {ch: run[pes_name, f"digitizers.{ch}.raw.samples"].select_trains(by_id[tids]).ndarray()
               for ch in channels}
    pes_raw_t = {ch: run_test[pes_name, f"digitizers.{ch}.raw.samples"].select_trains(by_id[test_tids]).ndarray()
                   for ch in channels}

    # select test SPEC data
    spec_raw_pe_t = run_test[spec_name, "data.photonEnergy"].select_trains(by_id[test_tids - spec_offset]).ndarray()
    spec_raw_int_t = run_test[spec_name, "data.intensityDistribution"].select_trains(by_id[test_tids - spec_offset]).ndarray()

    print("Data in memory.")

    # read the XGM information
    #xgm_pressure = run['SA3_XTD10_XGM/XGM/DOOCS', "pressure.pressureFiltered.value"].select_trains(by_id[tids]).ndarray()
    #xgm_pe =  run['SA3_XTD10_XGM/XGM/DOOCS:output', "data.intensitySa3TD"].select_trains(by_id[tids]).ndarray()
    #retvol_raw = run["SA3_XTD10_PES/MDL/DAQ_MPOD", "u212.value"].select_trains(by_id[tids]).ndarray()
    #retvol_raw_timestamp = run["SA3_XTD10_PES/MDL/DAQ_MPOD", "u212.timestamp"].select_trains(by_id[tids]).ndarray()

    xgm_flux =  run['SA3_XTD10_XGM/XGM/DOOCS:output', "data.intensitySa3TD"].select_trains(by_id[tids]).ndarray()[:, 0][:, np.newaxis]
    xgm_flux_t =  run_test['SA3_XTD10_XGM/XGM/DOOCS:output', "data.intensitySa3TD"].select_trains(by_id[test_tids]).ndarray()[:, 0][:, np.newaxis]

    print(f"Intensity in training: {np.mean(xgm_flux):.2e} +/- {np.std(xgm_flux):.2e}")
    print(f"Intensity in testing: {np.mean(xgm_flux_t):.2e} +/- {np.std(xgm_flux_t):.2e}")

    pressure = run["SA3_XTD10_PES/GAUGE/G30310F", "value"].select_trains(by_id[tids]).ndarray()
    pressure_t = run_test["SA3_XTD10_PES/GAUGE/G30310F", "value"].select_trains(by_id[test_tids]).ndarray()
    print(f"Pressure in training: {np.mean(pressure):.2e} +/- {np.std(pressure):.2e}")
    print(f"Pressure in testing: {np.mean(pressure_t):.2e} +/- {np.std(pressure_t):.2e}")

    voltage = run["SA3_XTD10_PES/MDL/DAQ_MPOD", "u212.value"].select_trains(by_id[tids]).ndarray()
    voltage_t = run_test["SA3_XTD10_PES/MDL/DAQ_MPOD", "u212.value"].select_trains(by_id[test_tids]).ndarray()
    print(f"Voltage in training: {np.mean(voltage):.2f} +/- {np.std(voltage):.2f}")
    print(f"Voltage in testing: {np.mean(voltage_t):.2f} +/- {np.std(voltage_t):.2f}")

    gas = get_gas(run, tids)
    gas_t = get_gas(run_test, test_tids)
    print(f"Gas in training: {gas}")
    print(f"Gas in testing: {gas_t}")

    t = list()
    t_names = list()

    model = Model(channels=channels, model_type=args.model_type)

    train_idx = np.isin(tids, train_tids) & (xgm_flux[:,0] > args.xgm_cut)
    # we just need this for training and we need to avoid copying it, which blows up the memoray usage
    for k in pes_raw.keys():
        pes_raw[k] = pes_raw[k][train_idx]

    model.debug_peak_finding(pes_raw, os.path.join(args.directory, "test_peak_finding.pdf"))
    if len(args.model) == 0:
        print("Fitting")
        start = time_ns()
        model.uniformize(xgm_flux[train_idx])
        model.fit(pes_raw,
                  #{k: v[train_idx]
                  #for k, v in pes_raw.items()},
                   spec_raw_int[train_idx],
                   spec_raw_pe[train_idx],
                   pulse_energy=xgm_flux[train_idx],
                   )
        t += [time_ns() - start]
        t_names += ["Fit"]

        print("Saving the model")
        start = time_ns()
        modelFilename = os.path.join(args.directory, "model.joblib")
        model.save(modelFilename)
        t += [time_ns() - start]
        t_names += ["Save"]
    else:
        print("Model has been given, so I will just load it.")
        modelFilename = args.model

    print("Loading the model")
    start = time_ns()
    model = Model.load(modelFilename)
    t += [time_ns() - start]
    t_names += ["Load"]

    # save PCA information
    pes_raw_select = model.x_select.transform(pes_raw, pulse_energy=xgm_flux[train_idx])
    ch = channels[0]
    idx = 0
    first = model.x_select.tof_start - model.x_select.delta_tof
    last = model.x_select.tof_start + model.x_select.delta_tof
    B, P, _ = pes_raw_select.shape
    pes_raw_select = pes_raw_select.reshape((B*P, -1))
    pca = PCA(None, whiten=True)
    pca.fit(pes_raw_select)
    df = pd.DataFrame(dict(variance_ratio=pca.explained_variance_ratio_,
                           n_comp=600*np.ones_like(pca.explained_variance_ratio_),
                           ))
    df.to_csv(os.path.join(args.directory, "pca_pes.csv"))

    pca_spec = PCA(None, whiten=True)
    pca_spec.fit(spec_raw_int[train_idx])
    df = pd.DataFrame(dict(variance_ratio=pca_spec.explained_variance_ratio_,
                           n_comp=20*np.ones_like(pca_spec.explained_variance_ratio_),
                           ))
    df.to_csv(os.path.join(args.directory, "pca_spec.csv"))

    # transfer function
    print(f"Resolution: {model.resolution:.2f} eV")

    df = pd.DataFrame(dict(wiener_energy=model.wiener_energy,
                           wiener_filter=model.wiener_filter,
                           impulse=model.impulse_response,
                           resolution=model.resolution*np.ones_like(model.wiener_energy)))
    df.to_csv(os.path.join(args.directory, "model.csv"))

    print("Check consistency")
    start = time_ns()
    Z = model.check_compatibility(pes_raw_t, pulse_energy=xgm_flux_t)
    print("Consistency check:", Z)
    Z = model.check_compatibility_per_channel(pes_raw_t)
    print("Consistency per channel:", Z)
    t += [time_ns() - start]
    t_names += ["Consistency"]

    # test
    print("Predict")
    start = time_ns()
    spec_pred = model.predict(pes_raw_t, pulse_energy=xgm_flux_t)
    spec_pred["deconvolved"] = model.deconvolve(spec_pred["expected"])
    t += [time_ns() - start]
    t_names += ["Predict"]

    print("Time taken in ms")
    df_time = pd.DataFrame(data=dict(time=t, name=t_names))
    df_time.time *= 1e-6
    print(df_time)

    print("Plotting")
    showSpec = False
    if len(args.model) == 0:
        showSpec = True
        # always smooth by 0.2, even if no smoothing was requested, for visualization
        smoother = HighResolutionSmoother(0.2)
        spec_smooth = smoother.fit_transform(spec_raw_int_t, energy=spec_raw_pe_t)
        # use whatever smoothing was used in the model for the calculations
        spec_data = model.preprocess_high_res(spec_raw_int_t)

        mse = np.mean((spec_data[:, np.newaxis, :] - spec_pred["expected"])**2, axis=(0, 1))
        var = np.std(spec_data, axis=0)**2
        r2 = 1 - mse/var
        print(f"MSE = {np.mean(mse):.2f}, var = {np.mean(var):.2f}, R^2 = {np.mean(r2):.2f}")

        # chi2 w.r.t XGM intensity
        chi2 = np.sum((spec_data[:, np.newaxis, :] - spec_pred["expected"])**2/(spec_pred["total_unc"]**2), axis=(-1, -2))
        ndof = spec_data.shape[1]
        print(f"Chi2 after PCA: {np.mean(chi2):.2f}, ndof: {ndof}, chi2/ndof: {np.mean(chi2/ndof):.2f}")

        spec_data_pca = model.y_model['pca'].transform(spec_data)
        unc2 = spec_pred["expected_pca_unc"]**2
        pca_var = (spec_pred["expected_pca"].std(axis=0, keepdims=True)**2).reshape(1, 1, -1)
        print("Expected pca std:", pca_var)
        chi2_prepca = np.sum((spec_data_pca[:, np.newaxis, :] - spec_pred["expected_pca"])**2/unc2, axis=(-1, -2))

        ndof_prepca = float(spec_data_pca.shape[-1])
        print(f"Chi2 before PCA: {np.mean(chi2_prepca):.2f}, ndof: {ndof_prepca}, chi2/ndof: {np.mean(chi2_prepca/ndof_prepca):.2f} +/- {np.std(chi2_prepca/ndof_prepca):.2f}")

        res_prepca = np.sum((spec_data_pca[:, np.newaxis, :] - spec_pred["expected_pca"])/spec_pred["expected_pca_unc"], axis=1)

        # rmse
        rmse = np.sqrt(np.mean((spec_data[:, np.newaxis, :] - spec_pred["expected"])**2, axis=(-1, -2)))
        nopca_unc = np.sqrt(np.mean(spec_pred["nopca_unc"]**2, axis=(-1, -2)))
        total_unc = np.sqrt(np.mean(spec_pred["total_unc"]**2, axis=(-1, -2)))
        median_unc = np.median(spec_pred["total_unc"], axis=(-1, -2))

        # sums
        first, last = model.get_low_resolution_range()
        pes_select_t = np.concatenate([-pes_raw_t[ch][:, first:last] for ch in pes_raw_t], axis=-1)
        sum_pes = np.sum(pes_select_t, axis=(-1))
        sum_spec = np.sum(spec_raw_int_t, axis=(-1))

        q = dict(chi2_prepca=chi2_prepca,
                 ndof=spec_data_pca.shape[-1]*np.ones_like(chi2_prepca),
                 xgm_flux_t=xgm_flux_t[:,0],
                 rmse=rmse,
                 nopca_unc=nopca_unc,
                 total_unc=total_unc,
                 median_unc=median_unc,
                 root_mean_squared_pca_unc=np.sqrt((spec_pred["expected_pca_unc"][:, 0, :]**2).mean(axis=-1)),
                 sum_pes=sum_pes,
                 sum_spec=sum_spec,
                 )
        q.update({f'res_prepca_{k}': res_prepca[:, k]
                  for k in range(res_prepca.shape[1])
                 }
                 )
        q.update({f'unc_prepca_{k}': spec_pred["expected_pca_unc"][:, 0, k]
                  for k in range(spec_pred["expected_pca_unc"].shape[-1])
                 }
                 )
        df = pd.DataFrame(q)
        df.to_csv(os.path.join(args.directory, "quality.csv"))

    first, last = model.get_low_resolution_range()
    # plot
    high_int_idx = np.argsort(xgm_flux_t[:,0])
    for q in [10, 25, 50, 75, 100]:
        qi = int(len(high_int_idx)*(q/100.0))
        for idx in high_int_idx[qi-10:qi]:
            tid = test_tids[idx]
            save_result(os.path.join(args.directory, f"test_q{q}_{tid}.csv"),
                       {k: item[idx, 0, ...] if k != "pca"
                           else item[0, ...]
                           for k, item in spec_pred.items()},
                        spec_raw_int_t[idx, :],
                        spec_smooth[idx, :],
                        spec_raw_pe_t[idx, :],
                        intensity=xgm_flux_t[idx,0],
                        )
            save_pes_result(os.path.join(args.directory, f"test_q{q}_{tid}_pes.csv"),
                        pes={k: -item[idx, :]
                             for k, item in pes_raw_t.items()},
                        first=first,
                        last=last,
                        )

if __name__ == '__main__':
    main()

