#!/usr/bin/env python

import os
import re

from typing import Optional, Tuple, Dict

import matplotlib
matplotlib.use('Agg')
import pandas as pd
import numpy as np
import matplotlib.pyplot as plt
from matplotlib.gridspec import GridSpec
import seaborn as sns

SMALL_SIZE = 12
MEDIUM_SIZE = 22
BIGGER_SIZE = 26

plt.rc('font', size=BIGGER_SIZE)         # controls default text sizes
plt.rc('axes', titlesize=BIGGER_SIZE)    # fontsize of the axes title
plt.rc('axes', labelsize=BIGGER_SIZE)    # fontsize of the x and y labels
plt.rc('xtick', labelsize=BIGGER_SIZE)   # fontsize of the tick labels
plt.rc('ytick', labelsize=BIGGER_SIZE)   # fontsize of the tick labels
plt.rc('legend', fontsize=MEDIUM_SIZE)   # legend fontsize
plt.rc('figure', titlesize=BIGGER_SIZE)  # fontsize of the figure title

def plot_resolution(df: pd.DataFrame, filename: str):
    fig = plt.figure(figsize=(12, 8))
    gs = GridSpec(1, 1)
    ax = fig.add_subplot(gs[0, 0])
    ax.plot(df.number_channels, df.resolution, c='b', lw=3)
    ax.spines['top'].set_visible(False)
    ax.spines['right'].set_visible(False)
    ax.set(
           xlabel="Number of channels",
           ylabel="Average resolution [eV]",
           )
    plt.tight_layout()
    fig.savefig(filename)
    plt.close(fig)

def plot_unc(df: pd.DataFrame, filename: str):
    fig = plt.figure(figsize=(12, 8))
    gs = GridSpec(1, 1)
    ax = fig.add_subplot(gs[0, 0])
    #ax.plot(df.number_channels, 2*df.unc, c='tab:blue', lw=3, alpha=0.7, label="Avg. 95% CL uncertainty band")
    #ax.fill_between(df.number_channels, 2*df.unc - 2*df.delta_unc, 2*df.unc + 2*df.delta_unc, color='tab:blue', alpha=0.2)
    ax.errorbar(df.number_channels, 2*df.unc, yerr=2*df.delta_unc, color='tab:blue', alpha=0.5, marker='o', markersize=20, lw=3, linestyle='none', label="95% CL uncertainty band")
    ax.spines['top'].set_visible(False)
    ax.spines['right'].set_visible(False)
    ax.set(
           xlabel="Number of channels",
           ylabel="Grating spectrometer intensity [a.u.]",
           )
    #rax = ax.twinx()
    rax = ax
    #rax.plot(df.number_channels, df.rmse, c='tab:red', lw=3, alpha=0.7, label="Avg. root-mean-squared error")
    #rax.fill_between(df.number_channels, df.rmse - df.delta_rmse, df.rmse + df.delta_rmse, color='tab:red', alpha=0.2)
    rax.errorbar(df.number_channels, df.rmse, yerr=df.delta_rmse, color='tab:red', alpha=0.5, marker='^', markersize=20, lw=3, linestyle='none', label="Root-mean-squared error")
    #rax.spines['right'].set_color('tab:red')
    #rax.spines['top'].set_visible(False)
    #rax.tick_params(axis='y', colors='tab:red')
    #rax.set_ylabel("Root-mean-squared error [a.u.]")
    #rax.yaxis.label.set_color("tab:red")
    ax.legend(frameon=False)
    plt.tight_layout()
    fig.savefig(filename)
    plt.close(fig)

def plot_rmse(df: pd.DataFrame, filename: str):
    fig = plt.figure(figsize=(12, 8))
    gs = GridSpec(1, 1)
    ax = fig.add_subplot(gs[0, 0])
    ax.plot(df.number_channels, df.rmse, c='b', lw=3)
    ax.spines['top'].set_visible(False)
    ax.spines['right'].set_visible(False)
    ax.set(
           xlabel="Number of channels",
           ylabel="Root-mean-squared error [a.u.]",
           )
    plt.tight_layout()
    fig.savefig(filename)
    plt.close(fig)

if __name__ == '__main__':
    indir = 'p900331r69t70'
    df = pd.read_csv(f'{indir}/number_channel_effect.csv')

    plot_rmse(df, "rmse.pdf")
    plot_unc(df, "unc.pdf")
    plot_resolution(df, "resolution.pdf")

