#!/usr/bin/env python

import os
import re

from typing import Optional, Tuple, Dict

import matplotlib
matplotlib.use('Agg')
import pandas as pd
import numpy as np
import matplotlib.pyplot as plt
from matplotlib.gridspec import GridSpec
import seaborn as sns

SMALL_SIZE = 12
MEDIUM_SIZE = 22
BIGGER_SIZE = 26

plt.rc('font', size=BIGGER_SIZE)         # controls default text sizes
plt.rc('axes', titlesize=BIGGER_SIZE)    # fontsize of the axes title
plt.rc('axes', labelsize=BIGGER_SIZE)    # fontsize of the x and y labels
plt.rc('xtick', labelsize=BIGGER_SIZE)   # fontsize of the tick labels
plt.rc('ytick', labelsize=BIGGER_SIZE)   # fontsize of the tick labels
plt.rc('legend', fontsize=MEDIUM_SIZE)   # legend fontsize
plt.rc('figure', titlesize=BIGGER_SIZE)  # fontsize of the figure title

def plot_final(df: pd.DataFrame, smooth: bool, filename: str):
    fig = plt.figure(figsize=(12, 8))
    gs = GridSpec(1, 1)
    ax = fig.add_subplot(gs[0, 0])
    if smooth:
        ax.plot(df.energy, df.spec_smooth, c='b', lw=3, label="Grating spectrometer (smoothened)")
        Y = np.amax(df.spec_smooth)
    else:
        ax.plot(df.energy, df.spec, c='b', lw=3, label="Grating spectrometer")
        Y = np.amax(df.spec)
    ax.plot(df.energy, df.prediction, c='r', ls='--', lw=3, label="Prediction")
    ax.fill_between(df.energy, df.prediction - 1*df.unc, df.prediction + 1*df.unc, facecolor='gold', alpha=0.5, label="68% unc. (total)")
    ax.fill_between(df.energy, df.prediction - 1*df.unc_pca, df.prediction + 1*df.unc_pca, facecolor='magenta', alpha=0.5, label="68% unc. (PCA only)")
    ax.legend(frameon=False, borderaxespad=0, loc='upper left')
    ax.set_title(f"Beam intensity: {df.beam_intensity.iloc[0]:.1f} mJ", loc="left")
    ax.spines['top'].set_visible(False)
    ax.spines['right'].set_visible(False)
    ax.set(
           xlabel="Photon energy [eV]",
           ylabel="Intensity [a.u.]",
           ylim=(0, 1.2*Y))
    plt.tight_layout()
    fig.savefig(filename)
    plt.close(fig)

def plot_chi2(df: pd.DataFrame, filename: str):
    fig = plt.figure(figsize=(12, 8))
    gs = GridSpec(1, 1)
    ax = fig.add_subplot(gs[0, 0])
    sns.histplot(x=df.chi2_prepca/df.ndof.iloc[0], kde=True, linewidth=3, ax=ax)
    ax.set(title=f"",
           xlabel=r"$\chi^2/$ndof",
           ylabel="Counts [a.u.]",
           xlim=(0, 5),
           )
    fig.savefig(filename)
    plt.close(fig)

def plot_rmse(df: pd.DataFrame, filename: str):
    fig = plt.figure(figsize=(12, 8))
    gs = GridSpec(1, 1)
    ax = fig.add_subplot(gs[0, 0])
    sns.histplot(x=df.rmse, kde=True, linewidth=3, ax=ax)
    ax.set(title=f"",
           xlabel=r"Root-mean-square-error",
           ylabel="Counts [a.u.]",
           )
    fig.savefig(filename)
    plt.close(fig)

def plot_residue(df: pd.DataFrame, filename: str):
    cols = [k for k in df.columns if "res_prepca" in k]
    df_res = df.loc[:, cols]
    n_plots = len(df_res.columns)//10
    fig = plt.figure(figsize=(8*n_plots, 8))
    gs = GridSpec(1, n_plots)
    for i_plot in range(n_plots):
        ax = fig.add_subplot(gs[0, i_plot])
        sns.kdeplot(data={f"Dim. {k+1}": df_res.loc[:, cols[k]] for k in range(i_plot*10, i_plot*10 + 10)},
                    linewidth=3, ax=ax)
        ax.set(title=f"",
           xlabel=r"residue/uncertainty [a.u.]",
           ylabel="Counts [a.u.]",
           xlim=(-3, 3),
           )
        #ax.legend(frameon=False)
    fig.savefig(filename)
    plt.close(fig)

def plot_residue_corr(df: pd.DataFrame, filename: str):
    cols = [k for k in df.columns if "res_prepca" in k]
    df_res = df.loc[:, cols]
    df_res.columns = [re.match(r"res_prepca_([0-9]*)", k).groups()[0] for k in df_res.columns]
    fig = plt.figure(figsize=(8, 8))
    corr = df_res.corr()
    mask = np.triu(np.ones_like(corr, dtype=bool))
    cmap = sns.diverging_palette(230, 20, as_cmap=True)
    sns.heatmap(corr, mask=mask, cmap=cmap, center=0,
                square=True, linewidths=0.5, vmin=-1, vmax=1)
    fig.savefig(filename)
    plt.close(fig)

def plot_chi2_intensity(df: pd.DataFrame, filename: str):
    fig = plt.figure(figsize=(12, 8))
    gs = GridSpec(1, 1)
    ax = fig.add_subplot(gs[0, 0])
    #sns.kdeplot(x=df.chi2_prepca/df.ndof.iloc[0], y=df.xgm_flux_t*1e-3,
    #            fill=True,
    #            ax=ax)
    sns.scatterplot(x=df.chi2_prepca/df.ndof.iloc[0], y=df.xgm_flux_t*1e-3,
                    s=5,
                    #alpha=0.4,
                    c="tab:red",
                    #size=df.root_mean_squared_pca_unc,
                    #sizes=(20, 200),
                    ax=ax)
    ax = plt.gca()
    ax.set(title=f"",
           xlabel=r"$\chi^2/$ndof",
           ylabel="Beam intensity [mJ]",
           xlim=(0, 5),
           ylim=(0, df.xgm_flux_t.mean()*1e-3 + 3*df.xgm_flux_t.std()*1e-3)
           )
    ax.spines['top'].set_visible(False)
    ax.spines['right'].set_visible(False)
    plt.tight_layout()
    fig.savefig(filename)
    plt.close(fig)

def plot_pes_vs_spec_intensity(df: pd.DataFrame, filename: str):
    fig = plt.figure(figsize=(12, 8))
    gs = GridSpec(1, 1)
    ax = fig.add_subplot(gs[0, 0])
    sns.scatterplot(x=df.sum_spec, y=df.sum_pes,
                    s=20,
                    alpha=0.4,
                    c="tab:red",
                    #size=df.root_mean_squared_pca_unc,
                    #sizes=(20, 200),
                    ax=ax)
    ax = plt.gca()
    ax.set(title=f"",
           xlabel=r"Total grating spec. intensity [a.u.]",
           ylabel="Total PES intensity [a.u.]",
           )
    ax.spines['top'].set_visible(False)
    ax.spines['right'].set_visible(False)
    plt.tight_layout()
    fig.savefig(filename)
    plt.close(fig)

def plot_xgm_vs_spec_intensity(df: pd.DataFrame, filename: str):
    fig = plt.figure(figsize=(12, 8))
    gs = GridSpec(1, 1)
    ax = fig.add_subplot(gs[0, 0])
    sns.scatterplot(x=df.sum_spec, y=df.xgm_flux_t*1e-3,
                    s=20,
                    alpha=0.4,
                    c="tab:red",
                    #size=df.root_mean_squared_pca_unc,
                    #sizes=(20, 200),
                    ax=ax)
    ax = plt.gca()
    ax.set(title=f"",
           xlabel=r"Total grating spec. intensity [a.u.]",
           ylabel="XGM pulse intensity [mJ]",
           )
    ax.spines['top'].set_visible(False)
    ax.spines['right'].set_visible(False)
    plt.tight_layout()
    fig.savefig(filename)
    plt.close(fig)

def plot_rmse_intensity(df: pd.DataFrame, filename: str):
    fig = plt.figure(figsize=(12, 8))
    gs = GridSpec(1, 1)
    ax = fig.add_subplot(gs[0, 0])
    sns.kdeplot(x=df.rmse, y=df.xgm_flux_t*1e-3,
                fill=True,
                ax=ax)
    sns.scatterplot(x=df.rmse, y=df.xgm_flux_t*1e-3,
                    s=5,
                    alpha=0.4,
                    c="tab:red",
                    #size=df.root_mean_squared_pca_unc,
                    #sizes=(20, 200),
                    ax=ax)
    ax = plt.gca()
    ax.set(title=f"",
           xlabel=r"Root-mean-squared error [a.u.]",
           ylabel="Beam intensity [mJ]",
           ylim=(0, df.xgm_flux_t.mean()*1e-3 + 3*df.xgm_flux_t.std()*1e-3)
           )
    ax.spines['top'].set_visible(False)
    ax.spines['right'].set_visible(False)
    plt.tight_layout()
    fig.savefig(filename)
    plt.close(fig)

def plot_unc_intensity(df: pd.DataFrame, filename: str):
    fig = plt.figure(figsize=(12, 8))
    gs = GridSpec(1, 1)
    ax = fig.add_subplot(gs[0, 0])
    sns.kdeplot(x=df.total_unc, y=df.xgm_flux_t*1e-3,
                fill=True,
                ax=ax)
    sns.scatterplot(x=df.total_unc, y=df.xgm_flux_t*1e-3,
                    s=5,
                    alpha=0.4,
                    c="tab:red",
                    #size=df.root_mean_squared_pca_unc,
                    #sizes=(20, 200),
                    ax=ax)
    ax = plt.gca()
    ax.set(title=f"",
           xlabel=r"Uncertainty [a.u.]",
           ylabel="Beam intensity [mJ]",
           ylim=(0, df.xgm_flux_t.mean()*1e-3 + 3*df.xgm_flux_t.std()*1e-3)
           )
    ax.spines['top'].set_visible(False)
    ax.spines['right'].set_visible(False)
    plt.tight_layout()
    fig.savefig(filename)
    plt.close(fig)

def plot_unc_rmse(df: pd.DataFrame, filename: str):
    fig = plt.figure(figsize=(12, 8))
    gs = GridSpec(1, 1)
    ax = fig.add_subplot(gs[0, 0])
    sns.kdeplot(x=2*df.total_unc, y=df.rmse,
                fill=True,
                ax=ax)
    sns.scatterplot(x=2*df.total_unc, y=df.rmse,
                    s=5,
                    alpha=0.4,
                    c="tab:red",
                    #size=df.root_mean_squared_pca_unc,
                    #sizes=(20, 200),
                    ax=ax)
    ax = plt.gca()
    ax.set(title=f"",
           xlabel=r"Root-mean-squared unc. [a.u.]",
           ylabel="Root-mean-squared error [a.u.]",
           )
    ax.spines['top'].set_visible(False)
    ax.spines['right'].set_visible(False)
    plt.tight_layout()
    fig.savefig(filename)
    plt.close(fig)

def pca_variance_plot(df: pd.DataFrame, filename: str, max_comp_frac: float=0.99):
    """
    Plot variance contribution.

    Args:
      filename: Output file name.
      variance_ratio: Contribution of each component's variance.

    """
    fig = plt.figure(figsize=(12, 8))
    gs = GridSpec(1, 1)
    ax = fig.add_subplot(gs[0, 0])
    c = np.cumsum(df.variance_ratio)
    n_comp = int(df.n_comp.iloc[0])
    ax.bar(1+np.arange(len(df.variance_ratio)), df.variance_ratio*100, color='tab:red', alpha=0.3, label="Per component")
    ax.plot(1+np.arange(len(df.variance_ratio)), c*100, c='tab:blue', lw=5, label="Cumulative")
    ax.plot([n_comp, n_comp], [0, c[n_comp]*100], lw=3, ls='--', c='m', label="Components kept")
    ax.plot([0, n_comp], [c[n_comp]*100, c[n_comp]*100], lw=3, ls='--', c='m')
    ax.legend(frameon=False)
    print(f"PCA plot: total n. components: {len(df.variance_ratio)}")
    x_max = np.where(c > max_comp_frac)[0][0]
    print(f"Fraction of variance: {c[n_comp]}")
    ax.set_yscale('log')
    ax.set(title=f"",
           xlabel="Component",
           ylabel="Variance contribution [%]",
           xlim=(1, x_max),
           ylim=(0.01, 100))
    ax.spines['top'].set_visible(False)
    ax.spines['right'].set_visible(False)
    plt.tight_layout()
    fig.savefig(filename)
    plt.close(fig)

def moving_average(a, n=3):
    ret = np.cumsum(a)
    ret[n:] = ret[n:] - ret[:-n]
    return ret[n - 1:] / n

def plot_impulse(df: pd.DataFrame, filename: str):
    """
    Plot variance contribution.

    Args:
      filename: Output file name.
      variance_ratio: Contribution of each component's variance.

    """
    fig = plt.figure(figsize=(12, 8))
    gs = GridSpec(1, 1)
    ax = fig.add_subplot(gs[0, 0])
    x = df.wiener_energy.to_numpy()
    y = np.absolute(df.impulse.to_numpy())
    #x_new = np.linspace(-6, 6, 601)
    #spl = make_interp_spline(x, np.log10(y), k=3)
    #y_new = np.power(10, spl(x_new))
    x_new = moving_average(x, n=5)
    y_new = moving_average(y, n=5)
    sel = (x_new >= -5.1) & (x_new <= 5.1)
    ax.plot(x_new[sel], y_new[sel], c='tab:blue', lw=3)
    ax.set_yscale('log')
    ax.set(title=f"",
           xlabel="Energy [eV]",
           ylim=(1e-4, 0.4),
           ylabel="Response [a.u.]",
           )
    ax.spines['top'].set_visible(False)
    ax.spines['right'].set_visible(False)
    plt.tight_layout()
    fig.savefig(filename)
    plt.close(fig)

def plot_wiener(df: pd.DataFrame, filename: str):
    """
    Plot variance contribution.

    Args:
      filename: Output file name.
      variance_ratio: Contribution of each component's variance.

    """
    fig = plt.figure(figsize=(12, 8))
    gs = GridSpec(1, 1)
    ax = fig.add_subplot(gs[0, 0])
    ax.plot(df.wiener_energy, np.absolute(df.wiener_filter), c='tab:blue', lw=3)
    ax.set_yscale('log')
    ax.set(title=f"",
           xlabel="Energy [eV]",
           ylim=(1e-3, 1),
           ylabel="Response [a.u.]",
           )
    ax.spines['top'].set_visible(False)
    ax.spines['right'].set_visible(False)
    plt.tight_layout()
    fig.savefig(filename)
    plt.close(fig)

def plot_pes(df: pd.DataFrame,
             channel: Dict[str, int],
             filename: str,
             fast_range: Optional[Tuple[int, int]]=None,
             Ne1s: Optional[Tuple[int, int]]=None,
             label: Optional[Dict[str, str]]=None,
             refs: Optional[Dict[str, Dict[int, float]]]=None,
             counts_to_mv: Optional[float]=None,
             tof: bool=False,
             ):
    """
    Plot low-resolution spectrum.

    Args:
      df: Data.
      channel: Prompt position per channel.
      filename: Output file name.
      fast_range: Range of fast electrons.
      Ne1s: range of Ne transitions.
      label: Labels per channel.
      refs: Reference.
      counts_to_mv: Multiplicative factor to convert counts to mV.
      tof: Show the tof or the energy?

    """
    fig = plt.figure(figsize=(12, 8))
    gs = GridSpec(1, 1)
    ax = fig.add_subplot(gs[0, 0])
    first, last = df.loc[:, 'first'].iloc[0], df.loc[:, 'last'].iloc[0]
    first = first+220
    last = last-270
    print("Range:", first, last)
    sel = (df.bin >= first) & (df.bin < last)
    x = df.loc[sel, "bin"].to_numpy()
    col = dict()
    colors = ["tab:red", "tab:blue"]
    p = list()
    # plot each channel
    for ich, ch in enumerate(channel.keys()):
        if label is None:
            sch = ch.replace('_', '')[-2:]
        else:
            sch = label[ch]
        y = df.loc[sel, ch].to_numpy().astype(np.float32)
        if counts_to_mv is not None:
            y *= counts_to_mv
        c = colors[ich]
        col[ch] = c
        p += [ax.plot(x, y, lw=2, c=c, label=sch)]
    ax.set(title=f"",
           ylim=(0, None),
           #xlabel="Time-of-flight index",
           xlabel="Samples",
           ylabel="Counts [a.u.]" if counts_to_mv is None else "Digitizer reading [mV]")
    ax.spines['top'].set_visible(False)
    ax.spines['right'].set_visible(False)
    minY, maxY = ax.get_ylim()
    # show reference energy lines
    if refs is not None:
        for ich, ch in enumerate(channel.keys()):
            for tof_, energy in refs[ch].items():
                ax.axvline(tof_, 0, 0.5 + ich*0.17, ls='-.', lw=1, c=col[ch])
                ax.text(tof_-1, (0.51 + ich*0.18)*maxY, f"{energy} eV", fontsize=14, rotation="vertical", color=col[ch])
    # show prompt line
    for ch, prompt in channel.items():
        ax.axvline(x=prompt, ls='--', lw=1, c=col[ch])
        ax.text(prompt-3, 0.5*maxY, "Prompt", fontsize=16, rotation="vertical", color=col[ch])
    # show the fast electrons range
    if fast_range is not None:
        x1, x2 = fast_range
        xtext = int(x1 + (x2 - x1)*0.3)
        ytext = 0.9*maxY
        ax.fill_between([x1, x2], minY, maxY, alpha=0.2, facecolor="tab:olive")
        ax.text(xtext, ytext, "Valence", fontsize=18, fontweight='bold')
        ax.text(xtext, ytext-0.05*maxY, "Auger", fontsize=18, fontweight='bold')
    # show the Ne 1s range
    if Ne1s is not None:
        x1, x2 = Ne1s
        xtext = int(x1 + (x2 - x1)*0.3)
        ytext = 0.9*maxY
        ax.fill_between([x1, x2], minY, maxY, alpha=0.2, facecolor="tab:cyan")
        ax.text(xtext, ytext, "Ne 1s", fontsize=22, fontweight='bold')
    ns_per_sample = 0.5
    cax = dict()
    def f_(ch):
        return (lambda kk: (np.array(kk) - int(channel[ch]))*ns_per_sample)
    def i_(ch):
        return (lambda kk: np.array(kk)/ns_per_sample + int(channel[ch]))
    RV = 108
    BE = 870
    PR = -9.02583275 + 0.99642416*RV + 14.0
    PL = 3.09786704e+04 - 3.85541614e+01*RV + (8.13753876e-03)*(RV**2)
    def f_to_ev(ch):
        return (lambda kk: PR + PL/((np.array(kk)*ns_per_sample - int(channel[ch])*ns_per_sample)**2) + BE)
    def i_to_ev(ch):
        return (lambda kk: np.sqrt(PL/(kk - PR - BE))/ns_per_sample + int(channel[ch]))
    if tof:
        forward_ = {ch: f_(ch) for ch in channel}
        inverse_ = {ch: i_(ch) for ch in channel}
    else:
        forward_ = {ch: f_to_ev(ch) for ch in channel}
        inverse_ = {ch: i_to_ev(ch) for ch in channel}

    for ich, (ch, prompt) in enumerate(channel.items()):
        cax[ch] = ax.secondary_xaxis(1.0+0.07*ich, functions=(forward_[ch], inverse_[ch]))
        #cax[ch].spines['left'].set_visible(False)
        cax[ch].spines['top'].set_position(('outward', 10))
        cax[ch].spines['top'].set_color(col[ch])
        cax[ch].tick_params(axis='x', colors=col[ch], labelsize=16)
        if ich == len(channel)-1:
            if tof:
                cax[ch].set_xlabel('Time-of-flight [ns]', fontsize=16)
            else:
                cax[ch].set_xlabel('Calibrated energy [eV]', fontsize=16)
            #cax[ch].xaxis.label.set_color(col[ch])
            #cax[ch].title.set_color(col[ch])
        if not tof:
            #locator = matplotlib.ticker.MultipleLocator(2)
            #locator.view_limits(1000, 1010)
            #cax[ch].xaxis.set_major_locator(locator)
            cax[ch].set_xticks(np.arange(995, 1010, 5))
    ax.legend(frameon=False, loc='center')
    plt.tight_layout()
    fig.savefig(filename)
    plt.close(fig)

if __name__ == '__main__':
    indir = 'p900331r69t70'
    channel = {'channel_4_A': 2639,
               'channel_3_B': 2646,
              }
    label = {'channel_4_A': r'22.5$^\circ$',
               'channel_3_B': r'225$^\circ$',
              }
    Ne1s = (2710, 2742)
    fast_range = (2650, 2670)
    refs={'channel_4_A': {2716:1002.5, 2722:997.5},
          'channel_3_B': {2723:1002.5, 2729:997.5}
          }
    counts_to_mv = 40.0/100.0
    #channel = 'sum'
    #for fname in os.listdir(indir):
    #    if re.match(r'test_q100_[0-9]*\.csv', fname):
    #        fname = fname[:-4]
    #        print(f"Plotting {fname}")
    #        plot_final(pd.read_csv(f'{indir}/{fname}.csv'), f'{fname}.pdf')
    #        plot_pes(pd.read_csv(f'{indir}/{fname}_pes.csv'), channel, f'{fname}_pes.pdf')

    for fname in ('test_q100_1724098413', 'test_q100_1724098596', 'test_q50_1724099445'):
        plot_final(pd.read_csv(f'{indir}/{fname}.csv'), smooth=True, filename=f'{fname}_smooth.pdf')
        plot_final(pd.read_csv(f'{indir}/{fname}.csv'), smooth=False, filename=f'{fname}.pdf')
        plot_pes(pd.read_csv(f'{indir}/{fname}_pes.csv'), channel, f'{fname}_pes.pdf',
                 fast_range=fast_range, Ne1s=Ne1s, label=label, refs=refs,
                 counts_to_mv=counts_to_mv)

    plot_pes_vs_spec_intensity(pd.read_csv(f'{indir}/quality.csv'), f'pes_vs_spec_intensity.pdf')
    plot_xgm_vs_spec_intensity(pd.read_csv(f'{indir}/quality.csv'), f'xgm_vs_spec_intensity.pdf')

    plot_chi2(pd.read_csv(f'{indir}/quality.csv'), f'chi2_prepca.pdf')
    plot_chi2_intensity(pd.read_csv(f'{indir}/quality.csv'), f'intensity_vs_chi2_prepca.pdf')
    plot_unc_intensity(pd.read_csv(f'{indir}/quality.csv'), f'intensity_vs_unc.pdf')
    plot_unc_rmse(pd.read_csv(f'{indir}/quality.csv'), f'rmse_vs_unc.pdf')
    plot_rmse(pd.read_csv(f'{indir}/quality.csv'), f'rmse.pdf')
    plot_rmse_intensity(pd.read_csv(f'{indir}/quality.csv'), f'intensity_vs_rmse.pdf')

    plot_residue(pd.read_csv(f'{indir}/quality.csv'), f'residue.pdf')
    plot_residue_corr(pd.read_csv(f'{indir}/quality.csv'), f'residue_corr.pdf')

    df_model = pd.read_csv(f'{indir}/model.csv')
    df_model.impulse = df_model.impulse.str.replace('i','j').apply(lambda x: complex(x))
    df_model.wiener_filter = df_model.wiener_filter.str.replace('i','j').apply(lambda x: complex(x))
    plot_impulse(df_model, f'impulse.pdf')
    plot_wiener(df_model, f'wiener.pdf')

    pca_variance_plot(pd.read_csv(f'{indir}/pca_spec.csv'), f'pca_spec.pdf', max_comp_frac=0.99)
    pca_variance_plot(pd.read_csv(f'{indir}/pca_pes.csv'), f'pca_pes.pdf', max_comp_frac=0.95)

